// Convolution of a 2D RGB image with an RGB filter. The filter
// is generated procedurally by a shader and the convolution is
// computed using the FFT on the GPU.
//
// Usage: Provide image as command line argument, 
// press <SPACE> to switch between image and filtered image.
//
// This demo uses the following software packages:
// GLUT, GLEW, SOIL, gpu_convolution, load_shaders, Timer
//
// Code written by Jeppe Revall Frisvad, August 2009.
// Copyright (c) DTU Informatics 2009

#include <iostream>
#include <GL/glew.h>
#include <GL/glut.h>
#include "../SOIL/SOIL.h"
#include "load_shaders.h"
#include "Convolution.h"
#include "OSTimer.h"

using namespace std;

//////////////////////////////////////////////////////////////////////
// Globals
//////////////////////////////////////////////////////////////////////

namespace
{
  int winw = 512, winh = 512;
  const unsigned int FFT_SIZE = 512;
  unsigned int disp_list;
  unsigned int source_tex;
  unsigned int filter_tex;
  unsigned int filter_prog;
  unsigned int fbo;
  bool draw_convolution = false; 
  Convolution* convolution;

  GLcharARB* pos_vert =
    "varying vec2 pos;\n"
    "void main()\n"
    "{\n"
    "  pos = (gl_ModelViewMatrix*gl_Vertex).xy;\n"
    "  gl_Position = ftransform();\n"
    "}";

  // Filter: (1 - dist^2*scale)^3
  // normalization factor == scale*4.0/(M_PI*radius*radius) == 0.31830989
  GLcharARB* filter_frag =
    "varying vec2 pos;\n"
    "void main()\n"
    "{\n"
    "  float dist = 1.0 - dot(pos, pos)*16384.0;\n"
    "  gl_FragColor.rgb = vec3(dist < 0.0 ? 0.0 : 0.31830989*dist*dist*dist);\n"
    "}";
}
  
//////////////////////////////////////////////////////////////////////
// Draw functions
//////////////////////////////////////////////////////////////////////

void set_projection()
{
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(-1.0, 1.0, -1.0, 1.0, -1.0, 1.0);
  glMatrixMode(GL_MODELVIEW);
}

void draw_source()
{
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glClear(GL_COLOR_BUFFER_BIT);
  set_projection();
  glLoadIdentity();

  glBindTexture(GL_TEXTURE_RECTANGLE_ARB, source_tex);
  glCallList(disp_list);
}

void draw_filter()
{
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glClear(GL_COLOR_BUFFER_BIT);
  set_projection();
  glLoadIdentity();

  glBindTexture(GL_TEXTURE_RECTANGLE_ARB, filter_tex);
  glCallList(disp_list);
}

//////////////////////////////////////////////////////////////////////
// Initialization
//////////////////////////////////////////////////////////////////////

void GLEWinit()
{
  GLenum err = glewInit();
  if(err != GLEW_OK)
  {
    cerr << "Error loading GLEW: " << glewGetErrorString(err) << endl;
    exit(0);
  }
}

void GLinit()
{
  glEnable(GL_CULL_FACE);
}

void init_texture(unsigned int& tex)
{
  glGenTextures(1, &tex);
  glBindTexture(GL_TEXTURE_RECTANGLE_ARB, tex);
	glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGBA32F_ARB, FFT_SIZE, FFT_SIZE, 0, GL_RGB, GL_FLOAT, 0);        
}

void init_framebuffer()
{
  glGenFramebuffersEXT(1, &fbo);
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
  glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_RECTANGLE_ARB, filter_tex, 0);
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);  
}

void init_disp_list()
{
  disp_list = glGenLists(1);
  glNewList(disp_list, GL_COMPILE);

  glEnable(GL_TEXTURE_RECTANGLE_ARB);
  glColor3f(1.0f, 1.0f, 1.0f);
  glBegin(GL_POLYGON);
    glTexCoord2f(0.0f,  0.0f);
    glVertex2f( -1.0f, -1.0f);
    glTexCoord2f(winw,  0.0f);
    glVertex2f(  1.0f, -1.0f);
    glTexCoord2f(winw,  winh);
    glVertex2f(  1.0f,  1.0f);
    glTexCoord2f(0.0f,  winh);
    glVertex2f( -1.0f,  1.0f);
  glEnd();
  glDisable(GL_TEXTURE_RECTANGLE_ARB);

  glEndList();
}

void init_source(const char* filename)
{
  int width, height, channels;
  unsigned char* data = SOIL_load_image(filename, &width, &height, &channels, SOIL_LOAD_RGB);
  if(!data)
  {
    cerr << "Error: could not load source image" << endl;
    exit(0);
  }

  source_tex = SOIL_create_OGL_texture(
    data, width, height, channels, 
    SOIL_CREATE_NEW_ID, 
    SOIL_FLAG_INVERT_Y | SOIL_FLAG_POWER_OF_TWO | SOIL_FLAG_TEXTURE_RECTANGLE);
  if(!source_tex)
  {
    cerr << "Error: could not create texture from source image" << endl;
    exit(0);
  }
}

void init_filter()
{
  const float step = 1.0f/static_cast<float>(FFT_SIZE);

  init_texture(filter_tex);
  init_framebuffer();
  load_shaders(pos_vert, filter_frag, filter_prog);

  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
  glClear(GL_COLOR_BUFFER_BIT);

  glUseProgram(filter_prog);
  glBegin(GL_POLYGON);
    glVertex2f(-1.0f, -1.0f);
    glVertex2f( 1.0f, -1.0f);
    glVertex2f( 1.0f,  1.0f);
    glVertex2f(-1.0f,  1.0f);
  glEnd();
  glUseProgram(0);

  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
}

void init(const char* source_file)
{
  GLEWinit();
  GLinit();
  init_disp_list();
  init_source(source_file);
  init_filter();

  convolution = new Convolution(FFT_SIZE, FFT_SIZE, source_tex, filter_tex);
}

//////////////////////////////////////////////////////////////////////
// GLUT Callback Functions
//////////////////////////////////////////////////////////////////////

void display()
{
  glClear(GL_COLOR_BUFFER_BIT);
  glLoadIdentity();

  if(draw_convolution)
  {    
    Timer t;
    t.start();
    convolution->convolve();
    convolution->draw_output(); 
    t.stop();
    cout << "Convolution computed in " << t.get_time() << " seconds." << endl;
  }
  else
    draw_source();

  glutSwapBuffers();
}

void reshape(int width, int height)
{
  winw = width;
  winh = height;

  glViewport(0, 0, winw, winh);
}

void keyboard(unsigned char key, int x, int y)
{
  switch(key)
  {
  case ' ':
    draw_convolution = !draw_convolution;
    glutPostRedisplay();
    break;
  case 27:
    delete convolution;
    exit(0);
  }
}

//////////////////////////////////////////////////////////////////////
// Main
//////////////////////////////////////////////////////////////////////

int main(int argc, char** argv)
{
  glutInit(&argc, argv);
  glutInitWindowSize(winw, winh);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH | GLUT_ALPHA);
  glutCreateWindow("GPU convolution demo");
  glutDisplayFunc(display);
  glutReshapeFunc(reshape);
  glutKeyboardFunc(keyboard);

  init(argc > 1 ? argv[1] : "test.png");

  glutMainLoop();

  return 0;
}
